from collections import OrderedDict
from shutil import ExecError

import torch
from gymnasium import spaces
from torch import nn

from general_q.encoders.base import Batch, Encoder


class MultiBatch(Batch[OrderedDict, OrderedDict]):
    @classmethod
    def from_dict(cls, data: OrderedDict, encoder, func=lambda x: x):
        return cls(
            OrderedDict(
                (key, func(value))
                for key, value in data.items()
            ),
            encoder
        )

    def __getitem__(self, item):
        return OrderedDict(
            (k, v[item])
            for k, v in self.data.items()
        )

    def __setitem__(self, key, value):
        for k, v in value.items():
            self._data[k][key] = v


class DictEncoder(Encoder, nn.ModuleDict):
    def __init__(self, space: spaces.Dict, subencoder=lambda: None, embed_dim=256):
        super().__init__(space)
        for key, subspace in space.spaces.items():
            encoder = subencoder(subspace, name=key, embed_dim=embed_dim)

            assert isinstance(encoder, Encoder), f"{encoder} should inherit from {Encoder}"
            self[key] = encoder

    @staticmethod
    def supports(space: spaces.Space) -> bool:
        return isinstance(space, spaces.Dict)

    def prepare(self, sample):
        return MultiBatch(
            OrderedDict(
                (key, encoder.prepare(sample[key]))
                for key, encoder in self.items()
            ),
            self
        )

    def sample(self, batch_size=1):
        return MultiBatch.from_dict(self, self, lambda encoder: encoder.sample(batch_size))

    def forward(self, x):
        return torch.sum(
            torch.stack([encoder(x.data[key]) for key, encoder in self.items()], dim=-1),
            dim=-1,
        )

    def item(self, sample):
        return OrderedDict(
            (key, encoder.item(sample.data[key]))
            for key, encoder in self.items()
        )

    def atomic_encoders(self):
        for key, encoder in self.items():
            for sub_key, sub_encoder in encoder.atomic_encoders():
                yield (key, *sub_key), sub_encoder

class TupleEncoder(DictEncoder):
    def __init__(self, space: spaces.Tuple, embed_dim=256):
        space = spaces.Dict(enumerate(space.spaces))
        super().__init__(space, embed_dim)

    @staticmethod
    def supports(space: spaces.Space) -> bool:
        return isinstance(space, spaces.Tuple)

    def prepare(self, sample):
        return super().prepare(OrderedDict(enumerate(sample)))

    def item(self, sample):
        return tuple(super().item(sample).values())



from typing import Generic, TypeVar

from abc import ABC, abstractmethod

import numpy as np
import torch
from gymnasium import spaces

from general_q.encoders.base import Batch, Encoder, I


class TensorBatch(Batch[torch.Tensor, I], Generic[I]):
    @property
    def data(self):
        return self._data

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        self._data[key] = value.data


M = TypeVar("M", bound=torch.nn.Module)


class TensorEncoder(Encoder[TensorBatch, I], ABC, Generic[M, I]):
    def __init__(self, space: spaces.Space[I], embed_dim: int):
        super().__init__(space)
        self.dtype = torch.tensor(space.sample()).dtype
        self.encoder: M = self.make_encoder(space, embed_dim)

    @abstractmethod
    def make_encoder(self, space: spaces.Space[I], embed_dim: int) -> M:
        pass

    def prepare(self, sample: I) -> TensorBatch:
        return TensorBatch(
            torch.tensor(sample, dtype=self.dtype, device=self.device),
            self,
        )

    def sample(self, batch_size=1, minibatch_size=64) -> TensorBatch:
        # TODO experiment with different minibatch sizes
        samples: list[torch.Tensor] = []
        for i in range(0, batch_size, minibatch_size):
            minibatch = np.array([
                self.space.sample()
                for _ in range(min(batch_size - i, minibatch_size))
            ])

            samples.append(self.prepare(minibatch).data)

        return TensorBatch(torch.cat(samples), self)

    def forward(self, x: TensorBatch) -> torch.FloatTensor:
        return self.encoder(x.data)

    def item(self, x):
        return np.array(x.data.cpu())



import torch

from general_q.encoders import Encoder


class ReplayMemory:
    def __init__(
            self,
            observation_encoder: Encoder,
            action_encoder: Encoder,
            capacity: int,
            device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
    ):
        self.observations = observation_encoder.sample(capacity)
        self.actions = action_encoder.sample(capacity)
        self.rewards = torch.zeros((capacity,), dtype=torch.float32, device=device)
        self.terminations = torch.ones((capacity,), dtype=torch.bool, device=device)
        self.truncations = torch.ones((capacity,), dtype=torch.bool, device=device)

        self.capacity = capacity
        self.last = capacity - 1
        self.size = 0

    def push(
            self,
            new_observation,
            action,
            reward=0.0,
            termination=False,
            truncation=False,
    ) -> None:
        # TODO profile push
        self.last = (self.last + 1) % self.capacity

        self.observations[self.last] = \
            self.observations.encoder.prepare(new_observation)
        self.actions[self.last] = self.actions.encoder.prepare(action)
        self.rewards[self.last] = reward
        self.terminations[self.last] = termination
        self.truncations[self.last] = truncation

        self.size = min(self.size + 1, self.capacity)

    def valid_indices(self):
        is_valid = ~self.terminations & ~self.truncations
        is_valid[self.last] = False

        indices, = torch.where(is_valid)
        return indices

    def sample(self, batch_size: int):
        indices = self.valid_indices()
        indices = indices[torch.randperm(len(indices))]  # randomize indices
        indices = indices[:batch_size]

        return self[indices]

    def close(self):
        self.truncations[self.last] = True

    def __len__(self):
        return len(self.valid_indices())

    def __getitem__(self, item):
        # convert slices to indices
        item = torch.arange(self.size)[item]
        next_item = (item + 1) % self.capacity

        obs = self.observations[item]
        action = self.actions[next_item]
        reward = self.rewards[next_item]
        terminations = self.terminations[next_item]
        truncations = self.truncations[next_item]
        new_obs = self.observations[next_item]

        return obs, action, reward, terminations, truncations, new_obs



import numpy as np
import torch
from gymnasium import spaces
from torch import nn

from general_q.encoders.tensor_encoder import TensorEncoder


class ImageEncoder(TensorEncoder):
    def make_encoder(self, space, embed_dim):
        self.encoder = nn.Sequential(
            nn.Conv2d(3, 16, kernel_size=3),
            nn.MaxPool2d(2),
            nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size=3),
            nn.MaxPool2d(2),
            nn.Conv2d(32, 64, kernel_size=3),
            nn.MaxPool2d(2),
            nn.Flatten(start_dim=-3),
            nn.ReLU(),
        )

        with torch.no_grad():
            sample = self.sample(1)
            sample = sample.to(torch.float32) / 255.0 - 0.5
            _, n = self(sample).shape

        self.encoder.append(nn.Linear(n, embed_dim))

    @staticmethod
    def supports(space: spaces.Space) -> bool:
        if not (
                type(space) == spaces.Box
                and space.dtype == np.uint8
                and len(space.shape) == 3
                and space.low.min() == 0
                and space.high.max() == 255
        ):
            return False

        h, w, c = space.shape
        return c == 3 and h >= 30 and w >= 30

    def prepare(self, sample):
        # rearrange last three dimensions to be (c, h, w)
        sample = super().prepare(sample)
        sample._data = sample._data.transpose(-1, -3)
        return sample

    def forward(self, x):
        return self.encoder(x.data.to(torch.float32) / 255.0 - 0.5)

    def item(self, x):
        return x.data.transpose(-1, -3).cpu().numpy(dtype=np.uint8)



# type: ignore[attr-defined]
"""An easy to use library for general purpose reinforcement learning and experimentation"""

from . import agents, encoders, utils


def get_version() -> str:
    from importlib import metadata as importlib_metadata

    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


__version__: str = get_version()



from typing import Generic, Sequence

from abc import ABC, abstractmethod

import torch
from gymnasium import spaces
from torch import nn

from general_q.encoders.base import Encoder, I, T
from general_q.encoders.tensor_encoder import TensorBatch, TensorEncoder


class Discrete(Encoder[T, I], ABC, Generic[T, I]):
    @abstractmethod
    def all(self) -> tuple[T, torch.FloatTensor]:
        """Make a batch of all possible values of this encoder"""


class DiscreteEncoder(TensorEncoder[nn.Embedding, int], Discrete[TensorBatch, int]):
    @staticmethod
    def supports(space: spaces.Space) -> bool:
        return type(space) == spaces.Discrete

    @torch.no_grad()
    def make_encoder(self, space: spaces.Discrete, embed_dim) -> nn.Embedding:
        encoder = nn.Embedding(space.n, embed_dim)

        start = encoder.weight[None, 0]
        end = encoder.weight[None, -1]
        weights = torch.linspace(0, 1, encoder.num_embeddings).unsqueeze(1)
        encoder.weight.data = start * (1 - weights) + end * weights
        return encoder

    def all(self):
        items = TensorBatch(torch.arange(self.encoder.num_embeddings, device=self.device), self)
        return items, self(items)

    def item(self, sample):
        return sample.data.item()


class MultiDiscreteEncoder(TensorEncoder[nn.ModuleList, Sequence[int]]):
    @staticmethod
    def supports(space) -> bool:
        return type(space) == spaces.MultiDiscrete

    def make_encoder(self, space, embed_dim):
        return nn.ModuleList([
            DiscreteEncoder(spaces.Discrete(n), embed_dim)
            for n in space.nvec
        ])

    def forward(self, x):
        x = x.data
        return torch.cat(
            [encoder(x[..., i]) for i, encoder in enumerate(self.encoder)],
            dim=-1,
        )


class MultiBinaryEncoder(TensorEncoder[nn.Sequential, I]):
    @staticmethod
    def supports(space: spaces.Space) -> bool:
        return type(space) == spaces.MultiBinary

    def make_encoder(self, space, embed_dim):
        class Float(nn.Module):
            def forward(self, x):
                return x.float()

        return nn.Sequential(
            Float(),
            nn.Flatten(start_dim=-len(space.shape)),
            nn.Linear(space.n, embed_dim),
        )



from typing import Optional, Union

import pickle
import random
from abc import ABC
from pathlib import Path

from gymnasium import Space
from gymnasium.core import ActType, ObsType

NAMES_PATH = Path(__file__).parent / "names.txt"

with open(NAMES_PATH) as f:
    NAMES = f.read().splitlines()


class Agent(ABC):
    def __init__(
            self,
            action_space: Space[ActType],
            observation_space: Space[ObsType],
            name: Optional[str] = None,
    ) -> None:
        """
        Initialize the class.

        Args:
            action_space: The action space of the environment.
            observation_space: The observation space of the environment.
            name: The name of the agent.
        """
        super().__init__()

        if name is None:
            name = random.choice(NAMES)

        self.action_space = action_space
        self.observation_space = observation_space
        self.name = name

    def act(self, obs: ObsType) -> tuple[ActType, float]:
        """
        Act based on the observation.

        Args:
            obs (np.ndarray): Observation from the environment.

        Returns:
            The action to take and the value of the action.
        """
        return self.action_space.sample(), 0.0

    def reset(self) -> None:
        """
        Reset the state of the agent. And get ready for next episode.
        """

    def remember(
            self,
            new_observation: ObsType,
            action: Optional[ActType] = None,
            reward: float = 0.0,
            termination: bool = False,
            truncation: bool = False,
    ) -> None:
        """
        Remember the action taken.

        Args:
            new_observation: The observation after the action.
            action: The action taken.
            reward: The reward received.
            termination: Whether the episode is done.
            truncation: Whether the episode is truncated.
        """

    def learn(self) -> float:
        """
        Perform one training step.

        Returns:
            The loss value.
        """
        return 0.0

    def close(self) -> None:
        """Close the agent."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def __call__(self, obs: ObsType) -> ActType:
        """Call agent.act()"""
        return self.act(obs)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r})'

    def save_pretrained(self, path: Union[str, Path]):
        """
        Save the agent to the given path.

        Args:
            path: The path to save the agent to.
        """

        path = Path(path) / f"{self.name}.{self.__class__.__name__}"
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load_pretrained(
            path: Union[str, Path],
            raise_error: bool = True,
    ) -> Optional["Agent"]:
        """
        Load the agent from the given path.

        Args:
            path: The path to load the agent from.
            raise_error: Whether to raise an error if the agent could not be loaded.

        Returns:
            The loaded agent or None if the agent could not be loaded.
        """
        try:
            with open(path, "rb") as f:
                return pickle.load(f)
        except (FileNotFoundError, EOFError, pickle.UnpicklingError):
            if raise_error:
                raise



from typing import Callable

from gymnasium import spaces

from general_q.encoders.base import Encoder, I, T
from general_q.encoders.box_encoder import BoxEncoder
from general_q.encoders.composite_encoders import DictEncoder, TupleEncoder
from general_q.encoders.discrete_encoders import (
    DiscreteEncoder,
    MultiBinaryEncoder,
    MultiDiscreteEncoder,
)
from general_q.encoders.image_encoder import ImageEncoder


def auto_encoder(space: spaces.Space[I], embed_dim: int, *args, **kwargs) -> Encoder[T, I]:
    encoders: Callable[[], Encoder[T, I]] = [
        lambda: ImageEncoder(space, embed_dim=embed_dim),
        lambda: BoxEncoder(space, embed_dim=embed_dim),
        lambda: DiscreteEncoder(space, embed_dim=embed_dim),
        lambda: MultiDiscreteEncoder(space, embed_dim=embed_dim),
        lambda: MultiBinaryEncoder(space, embed_dim=embed_dim),
        lambda: DictEncoder(space, subencoder=auto_encoder, embed_dim=embed_dim),
        lambda: TupleEncoder(space, subencoder=auto_encoder, embed_dim=embed_dim),
    ]

    for make_funciton in encoders:
        try:
            return make_funciton()
        except AssertionError:
            pass

    raise ValueError(f"No encoder for {space}, consider using a different subencoder")


def main():
    from gymnasium.spaces import Box, Dict, Discrete, MultiBinary, MultiDiscrete

    space = Dict(
        {
            "ext_controller": MultiDiscrete([5, 2, 2]),
            "inner_state":    Dict(
                {
                    "charge":        Discrete(100),
                    "system_checks": MultiBinary(10),
                    "job_status":    Dict(
                        {
                            "task":     Discrete(5),
                            "progress": Box(low=0, high=100, shape=()),
                        }
                    ),
                }
            ),
        }
    )

    encoder = auto_encoder(space, 64)
    sample = space.sample()
    new_sample = encoder.prepare(sample).item()

    encoded = encoder(encoder.sample(10))
    print(encoded.shape)


if __name__ == "__main__":
    main()



from typing import Optional

import torch
from gymnasium.core import ActType, ObsType
from gymnasium.spaces import Space
from torch import nn, optim
from torch.nn import functional as F

from general_q.agents.base import Agent
from general_q.agents.replay_memory import ReplayMemory
from general_q.encoders import DiscreteEncoder, Encoder, auto_encoder


class InvalidMemoryState(Exception):
    pass



class GeneralQ(Agent, nn.Module):
    def __init__(
            self,
            action_space: Space[ActType],
            observation_space: Space[ObsType],
            name: Optional[str] = None,
            action_encoder=DiscreteEncoder,
            observation_encoder=auto_encoder,
            q: Optional[nn.Module] = None,
            embed_dim: int = 256,
            device: torch.device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu"
            ),
            lr: float = 1e-3,
            replay_memory_size: int = 4096,
    ) -> None:
        super().__init__(action_space, observation_space, name)

        self.q = q
        if self.q is None:
            self.q = nn.Sequential(
                nn.Linear(embed_dim, embed_dim),
                nn.GELU(),
                nn.Linear(embed_dim, embed_dim),
                nn.GELU(),
                nn.Linear(embed_dim, embed_dim),
                nn.GELU(),
                nn.Linear(embed_dim, embed_dim),
                nn.GELU(),
                nn.Linear(embed_dim, 1),
                nn.Flatten(-2),
            )

        self.action_encoder = action_encoder(action_space, embed_dim)
        assert isinstance(self.action_encoder, Encoder), f"{self.action_encoder} should inherit from {Encoder}"

        self.observation_encoder = observation_encoder(observation_space, embed_dim)
        assert isinstance(self.observation_encoder, Encoder), f"{self.observation_encoder} should inherit from {Encoder}"

        self.optimizer = optim.Adam(self.parameters(), lr=lr)
        self.to(device)

        self.gameplays = ReplayMemory(
            action_encoder=self.action_encoder,
            observation_encoder=self.observation_encoder,
            capacity=replay_memory_size,
        )

    @torch.no_grad()
    def act(self, obs: ObsType) -> tuple[ActType, float]:
        obs = self.observation_encoder.prepare(obs)
        obs_embeddings = self.observation_encoder(obs)
        actions, action_embeddings = self.action_encoder.all()

        values = self.q(obs_embeddings + action_embeddings)
        best_value, best_action = values.max(dim=0)
        action = actions[best_action].item()

        return action, best_value.item()

    def remember(
            self,
            new_observation,
            action=None,
            reward=0,
            termination=False,
            truncation=False,
    ) -> None:
        if action is None:
            action = self.action_space.sample()
            if self.gameplays.size != 0:
                _, _, _, term, trunc, _ = self.gameplays[self.gameplays.last]
                state = "terminal" if term else "truncated" if trunc else ""

                if state:
                    raise InvalidMemoryState(
                        f"The last memory state was {state}, but no action was provided. "
                        f"Did you forget to call `agent.reset()` before `agent.remember(obs)`?"
                    )

        self.gameplays.push(
            new_observation,
            action,
            reward,
            termination,
            truncation,
        )

    def learn(self, batch_size=512) -> float:
        """
        Learn from the gameplays.

        Args:
            batch_size: The number of gameplay steps to learn from.

        Returns:
            The loss value.
        """

        batch_size = min(batch_size, len(self.gameplays))
        if batch_size == 0:
            return 0.0

        (
            observations,
            actions,
            rewards,
            terminations,
            _truncations,
            next_observations,
        ) = self.gameplays.sample(batch_size)

        next_observations = self.observation_encoder(next_observations)[:, None]  # [batch_size, 1, emb]
        _, action_embeddings = self.action_encoder.all()  # [n, emb]
        next_qs = self.q(next_observations + action_embeddings)  # [batch_size, n]
        best_q = next_qs.max(dim=1).values  # [batch_size]
        q_target = best_q * ~terminations + rewards  # [batch_size]

        observations = self.observation_encoder(observations)  # [batch_size, emb]
        actions = self.action_encoder(actions)  # [batch_size, emb]
        qs = self.q(observations + actions)  # [batch_size]
        loss = F.mse_loss(qs, q_target)

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return loss.item()

    def reset(self):
        self.gameplays.close()



import math

import numpy as np
from gymnasium import spaces
from torch import nn

from general_q.encoders.tensor_encoder import TensorEncoder


class BoxEncoder(TensorEncoder):
    def make_encoder(self, space: spaces.Box, embed_dim=256):
        class Unsqueeze(nn.Module):
            def forward(self, x):
                return x.unsqueeze(-1)

        preprocess_layer = (
            Unsqueeze()
            if space.shape == ()
            else nn.Flatten(start_dim=-len(space.shape))
        )
        return nn.Sequential(
            preprocess_layer,
            nn.Linear(math.prod(space.shape), embed_dim),
        )

    @classmethod
    def supports(cls, space: spaces.Space) -> bool:
        return type(space) == spaces.Box and space.dtype in [
            np.float16,
            np.float32,
            np.float64,
        ]



from typing import Callable, Generic, TypeVar

from abc import ABC, abstractmethod

import torch
from gymnasium import spaces
from torch import nn

T = TypeVar("T")
I = TypeVar("I")


class Batch(ABC, Generic[T, I]):
    def __init__(self, data: T, encoder) -> None:
        self._data = data
        self.encoder = encoder

    @property
    def data(self):
        return self._data

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def __setitem__(self, item, value) -> None:
        pass

    def item(self):
        return self.encoder.item(self)


class Encoder(nn.Module, ABC, Generic[T, I]):
    def __init__(self, space: spaces.Space[I]) -> None:
        assert self.supports(
            space
        ), f"{self.__class__.__name__} does not support {space}"

        super().__init__()
        self.space = space

        # phantom data to track device
        self.register_buffer("_phantom", torch.empty((0,)))

    __call__: Callable[["Encoder", T], torch.FloatTensor]

    @property
    def device(self) -> torch.device:
        return self._phantom.device

    @staticmethod
    def supports(space: spaces.Space) -> bool:
        """Returns whether this encoder supports the given space."""
        return False

    @abstractmethod
    def prepare(self, sample: I) -> T:
        """Prepare sample for forward pass."""

    @abstractmethod
    def sample(self, batch_size: int = 1) -> T:
        """Sample a batch of samples from the space."""

    @abstractmethod
    def forward(self, x: T) -> torch.FloatTensor:
        """Encode the given sample."""

    @abstractmethod
    def item(self, x):
        """Convert a tensor to a sample."""
    
    def atomic_encoders(self):
        yield (), self



from general_q.agents.base import NAMES, Agent
from general_q.agents.general_q import GeneralQ
from general_q.agents.replay_memory import ReplayMemory

__all__ = [
    'Agent',
    'NAMES',
    'GeneralQ',
    'ReplayMemory',
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['general_q', 'general_q.agents', 'general_q.decoders', 'general_q.encoders']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'general-q',
    'version': '0.1.0',
    'description': 'An easy to use library for general purpose reinforcement learning and experimentation',
    'long_description': '# general_q\n\n<div align="center">\n\n[![Build status](https://github.com/khoda81/general_q/workflows/build/badge.svg?branch=master&event=push)](https://github.com/khoda81/general_q/workflows/build/badge.svg?branch=master&event=push)\n[![Python Version](https://img.shields.io/pypi/pyversions/general_q.svg)](https://pypi.org/project/general_q/)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/khoda81/general_q/blob/master/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--versions-e10079.svg)](https://github.com/khoda81/general_q/releases)\n[![License](https://img.shields.io/github/license/khoda81/general_q)](https://github.com/khoda81/general_q/blob/master/LICENSE)\n![Coverage Report](assets/images/coverage.svg)\n\nAn easy to use library for general purpose reinforcement learning and experimentation\n\n</div>\n\n## Installation\n\n```bash\npip install -U general_q\n```\n\nor install with `Poetry`\n\n```bash\npoetry add general_q\n```\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/khoda81/general_q)](https://github.com/khoda81/general_q/blob/master/LICENSE)\n\nThis project is licensed under the terms of the `MIT` license.\nSee [LICENSE](https://github.com/khoda81/general_q/blob/master/LICENSE) for more details.\n\n## 📃 Citation\n\n```bibtex\n@misc{general_q,\n  author = {axiom},\n  title = {An easy to use library for general purpose reinforcement learning and experimentation},\n  year = {2022},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/khoda81/general_q}}\n}\n```\n\n## Credits [![🚀 Your next Python package needs a bleeding-edge project structure.](https://img.shields.io/badge/python--package--template-%F0%9F%9A%80-brightgreen)](https://github.com/TezRomacH/python-package-template)\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template)\n',
    'author': 'axiom',
    'author_email': '20.mahdikh.0@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/khoda81/general_q',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from itertools import count

from gymnasium import Env
from tqdm import tqdm, trange

from general_q.agents import Agent


def play(
        env: Env,
        agent: Agent,
        steps: int,
        train: bool = True,
        step_callback: callable = None,
        episode_callback: callable = None
) -> None:
    """
    Run a single episode of the environment.

    Args:
        env: The environment to run the episode in.
        agent: The agent to train.
        steps: The number of steps to run the environment for.
        train: Whether to train the agent.
        step_callback: A callback to call after each step.
        episode_callback: A callback to call after each episode.
    """
    if episode_callback is None:
        episode_callback = lambda *_, **__: None

    if step_callback is None:
        step_callback = lambda *_, **__: None

    try:
        max_step = env.spec.max_episode_steps
    except AttributeError:
        max_step = None

    with trange(steps) as step_pbar:
        while True:
            observation, info = env.reset()
            agent.reset()

            if train:
                agent.remember(observation)

            tot_rew = 0
            eps_loss = 0
            with tqdm(count(1), leave=False, total=max_step) as episode_pbar:
                for episode_step in episode_pbar:
                    action, value = agent(observation)
                    (
                        observation,
                        reward,
                        termination,
                        truncation,
                        info,
                    ) = env.step(action)

                    tot_rew += reward

                    step_callback(
                        step_pbar.n,
                        observation,
                        action,
                        reward,
                        termination,
                        truncation
                    )

                    description = ""
                    if train:
                        agent.remember(
                            observation,
                            action,
                            reward,
                            termination,
                            truncation
                        )

                        loss = agent.learn()
                        description += f"{loss=:6.3f}, "
                        eps_loss += loss

                    description += f"{reward=:6.2f}, {tot_rew=:6.2f}, {value=:6.2f}"
                    episode_pbar.set_description(description)
                    step_pbar.update()

                    if step_pbar.n >= steps:
                        return

                    if termination or truncation:
                        break

            episode_callback(
                step_pbar.n,
                episode_step,
                eps_loss / episode_step,
                tot_rew,
            )

            step_pbar.set_description(f"last_reward: {tot_rew:6.2f}")



from abc import ABC, abstractmethod
import gymnasium
from gymnasium import spaces
from torch import nn, distributions
import torch


class Distribution:
    pass


class Decoder(ABC, nn.Module):
    def __init__(self, space: gymnasium.Space):
        super().__init__()
        self.space = space

    @abstractmethod
    def forward(self, x: torch.Tensor) -> Distribution:
        """Decode the input into a distribution of self.space."""


class DiscreteDecoder(Decoder):
    def __init__(self, space: spaces.Discrete, embed_dim: int):
        super().__init__(space)
        self.decoder = nn.Linear(embed_dim, space.n)

    def forward(self, x: torch.Tensor) -> distributions.Categorical:
        return distributions.Categorical(logits=self.decoder(x))



from general_q.encoders.auto_encoder import auto_encoder
from general_q.encoders.base import Batch, Encoder
from general_q.encoders.box_encoder import BoxEncoder
from general_q.encoders.composite_encoders import DictEncoder, TupleEncoder
from general_q.encoders.discrete_encoders import (
    Discrete,
    DiscreteEncoder,
    MultiBinaryEncoder,
    MultiDiscreteEncoder,
)
from general_q.encoders.image_encoder import ImageEncoder
from general_q.encoders.tensor_encoder import TensorEncoder

__all__ = [
    'auto_encoder',
    'Batch',
    'BoxEncoder',
    'ImageEncoder',
    'DictEncoder',
    'Discrete',
    'DiscreteEncoder',
    'Encoder',
    'MultiBinaryEncoder',
    'MultiDiscreteEncoder',
    'TensorEncoder',
    'TupleEncoder',
]
